from __future__ import absolute_import

from celery import Celery

from omegaml.runtimes.jobproxy import OmegaJobProxy
from omegaml.runtimes.scriptproxy import OmegaScriptProxy
from omegaml.util import settings
import logging

logger = logging.getLogger(__name__)


class CeleryTask(object):
    """
    A thin wrapper for a Celery.Task object

    This is so that we can collect common delay arguments on the
    .task() call
    """

    def __init__(self, task, **kwargs):
        """

        Args:
            task (Celery.Task): the celery task object
            **kwargs (dict): optional, the kwargs to pass to apply_async
        """
        self.task = task
        self.kwargs = kwargs

    def apply_async(self, args=None, kwargs=None, *args_, **kwargs_):
        kwargs.update(self.kwargs)
        return self.task.apply_async(args=args, kwargs=kwargs, *args_)

    def delay(self, *args, **kwargs):
        """
        submit the task with args and kwargs to pass on

        This calls task.apply_async and passes on the self.kwargs.
        """
        return self.apply_async(args=args, kwargs=kwargs, **self.kwargs)

    def run(self, *args, **kwargs):
        return self.delay(*args, **kwargs)


class OmegaRuntime(object):
    """
    omegaml compute cluster gateway 
    """

    def __init__(self, omega, bucket=None, defaults=None, celeryconf=None):
        self.omega = omega
        defaults = defaults or settings()
        self.bucket = bucket
        self.pure_python = getattr(defaults, 'OMEGA_FORCE_PYTHON_CLIENT', False)
        self.pure_python = self.pure_python or self._client_is_pure_python()
        # initialize celery as a runtimes
        taskpkgs = defaults.OMEGA_CELERY_IMPORTS
        celeryconf = celeryconf or defaults.OMEGA_CELERY_CONFIG
        # ensure we use current value
        celeryconf['CELERY_ALWAYS_EAGER'] = bool(defaults.OMEGA_LOCAL_RUNTIME)
        self.celeryapp = Celery('omegaml')
        self.celeryapp.config_from_object(celeryconf)
        # needed to get it to actually load the tasks (???)
        # https://stackoverflow.com/a/35735471
        self.celeryapp.autodiscover_tasks(taskpkgs, force=True)
        self.celeryapp.finalize()
        # temporary requirements, use .require() to set
        self._require_kwargs = {}
        # fixed default arguments, use .require(always=True) to set
        self._task_default_kwargs = {}

    def __repr__(self):
        return 'OmegaRuntime({})'.format(self.omega.__repr__())

    @property
    def _common_kwargs(self):
        common = dict(pure_python=self.pure_python,
                      __bucket=self.bucket)
        common.update(self._task_default_kwargs)
        common.update(self._require_kwargs)
        return common

    def _client_is_pure_python(self):
        try:
            import pandas as pd
            import numpy as np
            import sklearn
        except Exception as e:
            logging.getLogger().info(e)
            return True
        else:
            return False

    def require(self, always=False, **kwargs):
        """
        specify requirements for the task execution

        Use this to specify resource or routing requirements on the next task
        call sent to the runtime. Any requirements will be reset after the
        call has been submitted.

        Args:
            always (bool): if True requirements will persist across task calls. defaults to False
            **kwargs: requirements specification that the runtime understands

        Usage:
            # celery runtime
            om.runtime.require(queue='gpu').model('foo').fit(...)

            # dask distributed runtime
            om.runtime.require(resource='gpu').model('foo').fit(...)

        Returns:
            self
        """
        if always:
            self._task_default_kwargs.update(kwargs)
        else:
            self._require_kwargs.update(kwargs)
        return self

    def model(self, modelname, require=None):
        """
        return a model for remote execution
        """
        from omegaml.runtimes.modelproxy import OmegaModelProxy
        self.require(require) if require else None
        return OmegaModelProxy(modelname, runtime=self)

    def job(self, jobname, require=None):
        """
        return a job for remote exeuction
        """
        self.require(require) if require else None
        return OmegaJobProxy(jobname, runtime=self)

    def script(self, scriptname):
        """
        return a script for remote execution
        """
        return OmegaScriptProxy(scriptname, runtime=self)

    def task(self, name, **kwargs):
        """
        retrieve the task function from the celery instance

        we do it like this so we can per-OmegaRuntime instance
        celery configurations (as opposed to using the default app's
        import, which seems to confuse celery)
        """
        kwargs.update(self._common_kwargs)
        taskfn = self.celeryapp.tasks.get(name)
        assert taskfn is not None, "cannot find task {name} in Celery runtime".format(**locals())
        task = CeleryTask(taskfn, **kwargs)
        self._require_kwargs = {}
        return task

    def settings(self, require=None):
        """
        return the runtimes's cluster settings
        """
        self.require(require) if require else None
        return self.task('omegaml.tasks.omega_settings').delay().get()

    def ping(self, require=None, *args, **kwargs):
        """
        ping the runtimes
        """
        self.require(require) if require else None
        return self.task('omegaml.tasks.omega_ping').delay(*args, **kwargs).get()



"""
Usage: om <command> [<action>] [<args> ...] [options]
       om (models|datasets|scripts|jobs) [<args> ...] [--raw] [--replace] [options]
       om runtime [<args> ...] [--async] [options]
       om cloud [<args> ...] [options]
       om shell [options]
       om [-h] [-hh] [--version] [--copyright]

omega|ml command line client

[usage:datasets]
[usage:models]
[usage:runtime]
[usage:scripts]
[usage:jobs]
[usage:cloud]
[usage:shell]

Options:
  -h --help          Show this screen
  -hh --help-ext     Show detailed descriptions
  --version          Show version.
  --loglevel=LEVEL   INFO,ERROR,DEBUG [default: INFO]
  --copyright        Show copyright
  --config=CONFIG    configuration file
  --bucket=BUCKET    the bucket to use
  --local-runtime    use local runtime

Options for datasets and models
  --raw  list Metadata objects instead of names only

[options:datasets]
[options:models]
[options:runtime]
[options:cloud]
[options:scripts]
[options:jobs]

[description:datasets]
[description:models]
[description:runtime]
[description:cloud]
[description:scripts]
[description:jobs]
"""
import sys

from omegaml import version
from omegaml.client.cli.catchall import GlobalCommand
from omegaml.client.cli.cloud import CloudCommand
from omegaml.client.cli.datasets import DatasetsCommand
from omegaml.client.cli.jobs import JobsCommand
from omegaml.client.cli.models import ModelsCommand
from omegaml.client.cli.runtime import RuntimeCommand
from omegaml.client.cli.scripts import ScriptsCommand
from omegaml.client.cli.shell import ShellCommand
from omegaml.client.docoptparser import DocoptParser


def main(argv=None, logger=None):
    # make sure cli sees current project
    sys.path.insert(0, '.')
    # use argv and logger for debugging and testing
    parser = DocoptParser(__doc__, [DatasetsCommand,
                                    ScriptsCommand,
                                    ModelsCommand,
                                    RuntimeCommand,
                                    CloudCommand,
                                    ShellCommand,
                                    JobsCommand,
                                    GlobalCommand],
                          argv=argv,
                          version=version,
                          logger=logger)
    parser.parse()
    parser.process()
    return parser

def climain():
    # entry point for distutils console_script
    main()

if __name__ == '__main__':
    main()



import numpy as np
from six import BytesIO

from omegaml.backends.basedata import BaseDataBackend


class NumpyNDArrayBackend(BaseDataBackend):
    """
    Store numpy NDArray of any shape or size

    The NDArray is serialized to a byte string and stored as a BLOB.
    Thus it can have arbitrary size and dimensions, ideal for image
    data and Tensors.
    """
    KIND = 'ndarray.bin'

    @classmethod
    def supports(self, obj, name, **kwargs):
        return isinstance(obj, np.ndarray)

    def put(self, obj, name, attributes=None, **kwargs):
        kind_meta = {
            'dtype': obj.dtype.name,
            'shape': obj.shape,
        }
        fout = self.data_store.fs.new_file(filename=name)
        buf = BytesIO()
        np.save(buf, obj)
        buf.seek(0)
        fout.write(buf)
        fout.close()
        return self.data_store.make_metadata(name, self.KIND, attributes=attributes,
                                             kind_meta=kind_meta,
                                             **kwargs).save()

    def get(self, name, version=-1, force_python=False, lazy=False, **kwargs):
        fin = self.data_store.fs.get_last_version(name)
        meta = self.data_store.metadata(name)
        dtype_name = meta.kind_meta['dtype']
        dtype = getattr(np, dtype_name, None)
        shape = meta.kind_meta['shape']
        buf = BytesIO()
        buf.write(fin.read())
        fin.close()
        buf.seek(0)
        return np.load(buf)



from omegaml.client.docoptparser import DocoptCommand
from omegaml.client.userconf import save_userconfig_from_apikey
from omegaml.client.util import get_omega
from omegaml.defaults import update_from_config


class CloudCommand(DocoptCommand):
    """
    Usage:
      om cloud login [<userid>] [<apikey>] [options]
      om cloud config [options]

    Options:
      --userid=USERID  the userid at hub.omegaml.io (see account profile)
      --apikey=APIKEY  the apikey at hub.omegaml.io (see account profile)
      --apiurl=URL     the cloud URL [default: https://hub.omegaml.io]
    """
    command = 'cloud'

    def login(self):
        userid = self.args.get('<userid>') or self.args.get('--userid')
        apikey = self.args.get('<apikey>') or self.args.get('--apikey')
        api_url = self.args.get('--apiurl')
        configfile = self.args.get('--config') or 'config.yml'
        if not userid:
            userid = input('Userid: ')
        if not apikey:
            apikey = input('Apikey: ')
        save_userconfig_from_apikey(configfile, userid, apikey, api_url=api_url)

    def config(self):
        om = get_omega(self.args)
        config_file = om.defaults.OMEGA_CONFIG_FILE
        if config_file is None:
            config_file = print("No configuration file identified, assuming defaults")
        # print config
        restapi_url = getattr(om.defaults, 'OMEGA_RESTAPI_URL', 'not configured')
        runtime_url = om.runtime.celeryapp.conf['BROKER_URL']
        userid = getattr(om.defaults, 'OMEGA_USERID', 'not configured')
        self.logger.info('Config file: {config_file}'.format(**locals()))
        self.logger.info('User id: {userid}'.format(**locals()))
        self.logger.info('REST API URL: {restapi_url}'.format(**locals()))
        self.logger.info('Runtime broker: {runtime_url}'.format(**locals()))





        









import os

import six
import yaml

from omegaml.client.userconf import get_omega_from_apikey
from omegaml.client.util import protected
from omegaml.omega import Omega as CoreOmega
from omegaml.runtimes import OmegaRuntime


class OmegaCloud(CoreOmega):
    """
    Client API to omegaml cloud

    Provides the following APIs:

    * :code:`datasets` - access to datasets stored in the cluster
    * :code:`models` - access to models stored in the cluster
    * :code:`runtimes` - access to the cluster compute resources
    * :code:`jobs` - access to jobs stored and executed in the cluster
    * :code:`scripts` - access to lambda modules stored and executed in the cluster

    """

    def __init__(self, auth=None, **kwargs):
        """
        Initialize the client API
        """
        super(OmegaCloud, self).__init__(**kwargs)
        self.runtime = OmegaCloudRuntime(self, defaults=self.defaults)
        self.auth = auth

    def _clone(self, **kwargs):
        kwargs.update(auth=self.auth)
        return super()._clone(**kwargs)

    def __repr__(self):
        return 'OmegaCloud(bucket={}, auth={})'.format(self.bucket or 'default', repr(self.auth))


class OmegaCloudRuntime(OmegaRuntime):
    """
    omegaml hosted compute cluster gateway
    """

    def __init__(self, omega, **kwargs):
        super().__init__(omega, **kwargs)
        self._auth_kwarg = protected('auth')

    def __repr__(self):
        return 'OmegaCloudRuntime(auth={})'.format(repr(self.omega.auth))

    @property
    def _common_kwargs(self):
        common = super()._common_kwargs
        common.update({self._auth_kwarg: self.auth_tuple})
        return common

    @property
    def auth_tuple(self):
        auth = self.omega.auth
        return auth.userid, auth.apikey, auth.qualifier


def setup(userid=None, apikey=None, api_url=None, qualifier=None, bucket=None):
    # from now on, link OmegaCloud implementation as the default
    import omegaml as om
    api_url = api_url or os.environ.get('OMEGA_RESTAPI_URL') or 'https://hub.omegaml.io'
    om.Omega = OmegaCloud
    om.setup = setup
    om.get_omega_for_task = lambda *args, **kwargs: setup(*args, **kwargs)
    om = get_omega_from_apikey(userid, apikey, api_url=api_url, qualifier=qualifier)
    return om[bucket]


def setup_from_config(config_file=None):
    from omegaml import _base_config
    config_file = config_file or _base_config.OMEGA_CONFIG_FILE
    if isinstance(config_file, six.string_types) and os.path.exists(config_file):
        with open(config_file, 'r') as fin:
            userconfig = yaml.safe_load(fin)
            try:
                omega = setup(userid=userconfig['OMEGA_USERID'],
                              apikey=userconfig['OMEGA_APIKEY'],
                              api_url=userconfig['OMEGA_RESTAPI_URL'])
            except:
                raise ValueError('Could not login using config file {}'.format(config_file))
            else:
                omega.defaults.OMEGA_CONFIG_FILE = config_file
            return omega
    raise SystemError('Config file {} does not exist'.format(config_file))



import warnings
from unittest import TestCase

import sys

import os


class OmegaTestMixin(object):
    def clean(self, bucket=None):
        om = self.om[bucket] if bucket is not None else self.om
        drop = om.models.drop
        [drop(m, force=True) for m in om.models]
        drop = om.datasets.drop
        [drop(m, force=True) for m in om.datasets]
        self.assertListEqual(om.datasets.list(), [])
        self.assertListEqual(om.models.list(), [])


def tf_in_eager_execution():
    # condition for unittest.skipIf decorator
    import tensorflow as tf
    return tf.executing_eagerly()

def tf_perhaps_eager_execution(*args, **kwargs):
    """
    test support to enable tf eager execution

    conditionally set eager execution if TF_EAGER env
    variable is yet to 1. else the eager state is not
    changed.
    """
    tf_eager_switch = os.environ.get('TF_EAGER', False)
    if int(tf_eager_switch):
        import tensorflow as tf
        try:
            tf.enable_eager_execution(*args, **kwargs)
            warnings.warn('TensorFlow eager execution enabled')
        except ValueError as e:
            warnings.warn(str(e))
    else:
        warnings.warn('TensorFlow eager execution not enabled TF_EAGER={tf_eager_switch}'.format(**locals()))







from unittest import TestCase

from sklearn.linear_model import LinearRegression

from omegaml import Omega
from omegaml.mixins.store.promotion import PromotionMixin
from omegaml.tests.util import OmegaTestMixin


class PromotionMixinTests(OmegaTestMixin, TestCase):
    def setUp(self):
        om = self.om = Omega()
        om.datasets.register_mixin(PromotionMixin)
        om.models.register_mixin(PromotionMixin)
        self.clean()
        self.clean('prod')

    def test_dataset_promotion(self):
        om = self.om
        prod = om['prod']
        om.datasets.put(['foo'], 'foo')
        # ensure dataset is in default bucket, not in prod
        self.assertIn('foo', om.datasets.list())
        self.assertNotIn('foo', prod.datasets.list())
        # promote to prod
        om.datasets.promote('foo', prod.datasets)
        self.assertIn('foo', prod.datasets.list())

    def test_model_promotion(self):
        om = self.om
        prod = om['prod']
        reg = LinearRegression()
        om.models.put(reg, 'mymodel')
        # ensure dataset is in default bucket, not in prod
        self.assertIn('mymodel', om.models.list())
        self.assertNotIn('mymodel', prod.models.list())
        # promote to prod
        om.models.promote('mymodel', prod.models)
        self.assertIn('mymodel', prod.models.list())



from omegaml.client.docoptparser import DocoptCommand
from omegaml.client.util import get_omega


class DatasetsCommand(DocoptCommand):
    """
    Usage:
      om datasets list [<pattern>] [--raw]
      om datasets put <path> <name> [--replace]
      om datasets get <name> <path>
      om datasets drop <name> [--force]
      om datasets metadata <name>
    """
    command = 'datasets'

    def list(self):
        om = get_omega(self.args)
        raw = self.args.get('--raw', False)
        pattern = self.args.get('<pattern>')
        entries = om.datasets.list(pattern=pattern, raw=raw)
        self.logger.info(entries)

    def put(self):
        om = get_omega(self.args)
        local = self.args['<path>']
        name = self.args['<name>']
        replace = self.args['--replace']
        if local.endswith('.csv'):
            import pandas as pd
            data = pd.read_csv(local)
            self.logger.info(om.datasets.put(data, name, append=not replace))
        else:
            self.logger.info(om.datasets.put(local, name, append=not replace))

    def get(self):
        om = get_omega(self.args)
        local = self.args['<path>']
        name = self.args['<name>']
        data = om.datasets.get(name).read()
        with open(local, 'wb') as fout:
            fout.write(data)
        self.logger.debug(local)

    def drop(self):
        om = get_omega(self.args)
        name = self.args['<name>']
        force = self.args['--force']
        om.datasets.drop(name, force=force)

    def metadata(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        self.logger.info(om.datasets.metadata(name).to_json())

    def plugins(self):
        om = get_omega(self.args)
        for kind, plugincls in om.defaults.OMEGA_STORE_BACKENDS.items():
            self.logger.info(kind, plugincls.__doc__)



"""
A parser/processor to simplify modular docopt cli
"""
import logging
import re
import sys
from textwrap import dedent

import os
from docopt import docopt, DocoptExit, DocoptLanguageError


class DocoptParser:
    """
    A simplified way to implement modular doctopt processors

    Why?
        To make cli implementation a breeze, perhaps even fun.

        doctopt is a great way to specify and parse command line arguments.
        However the developer is left to provide an interpretation and
        implementation of the parsed arguments. The implementation quickly
        becomes a large cascade of if/then/else: a bug feast waiting
        to happen and a maintenance nightmare.

        DocoptParser provides a straight-forward way to implement modular
        commands for processing, using docopt as the command line parser. In
        a nutshell, a command is a Python class that provides a method for
        each action to be performed, thus removing the need for large if/then
        constructs while modularising documentation and implementation of
        subcommands.

    Usage:
        ___doc___ = ''' Usage: pgm.py foo bar [--baz]

        [usage:foo]

        Options:
            -h | --help  this help text

        [options:foo]
        '''
        DocoptParser(__doc__, [FooCommand, ...]).parse().process()

        where FooCommand is

            class FooCommand(DocoptCommand):
                '''
                Usage: pgm.py foo bar [--baz]

                Options:
                    --baz  baz description
                '''
                command = 'foo'
                usage_header = 'Working with foo' # optional
                options_header = 'Options for foo' # optional

                def bar(self):
                    # the bar command, doctopt output is in self.args
                    # to write output, use self.logger, a Python logger
                    # which by default prints to stdout


        Upon the call to .process(), DocoptParser will find and call the
        FooCommmand.bar() method automatically. In bar() use self.args
        to get the parsed arguments as returned by docopt().

        DocoptParser will automatically replace [usage:foo] and [options:foo]
        with the FooCommand.__doc__ string, replacing the latter's Usage:
        with the usage_header and Options: with options_header

        The help text will thus be:

            Usage: pgm.py foo bar [--baz]

            Working with foo
                pgm.py foo bar [--baz]

            Options:
                -h | --help  this help text

            Options for foo
                --baz  baz description

    How to print output in a command:
        You should avoid using print() as it is not really a good way to
        write modular, debuggable code. If your cli is more than a few
        lines of code, it is better to use the built-in logger:

            (...)
            def bar(self):
                self.logger.info('this will go to stdout')

        To change the format of the output, create a logging.Logger and
        set whatever format you like, then use

            logger = ... # your logger setup
            DocoptParser(..., logger=logger)

        To change the loglevel, add --loglevel in options, pass INFO, DEBUG
        or ERROR

    Debugging:
        To debug the parser, set DOCOPT_DEBUG=1. This will print the parsed
        arguments in self.args, along with the command class and method chosen
        for execution on callling parser.process().

        If you don't see debug output despite specifying DOCOPT_DEBUG=1 the
        parser has not been able to parse the arguments and instead prints the
        help text. This means that your arguments don't match any of the
        specified arguments. Make sure that your top-level docstring supports
        all arguments and options.

        The command class is determined by a match to DocoptCommand.command,
        either of and in this order:
            1. a valid literal passed in argv
            2. the value of the <command> tag

        To change the <command> tag to something else, set parser.command_tag:

            parser = DocoptParser(..., command_tag='<dothis>')
    """

    def __init__(self, docs, commands, argv=None, version=None, logger=None,
                 command_tag=None):
        self.docs = docs
        self.commands = commands  # available command implementations
        self.command = None  # chosen command implementation, see get_command
        self.command_tag = command_tag or '<command>'
        self.argv = argv or sys.argv[1:]
        self.version = version
        self.args = None
        self._logger = logger
        self.apply_command_usage()

    @property
    def logger(self):
        if self._logger is None:
            self._logger = setup_console_logger()
            loglevel = self.args.get('--loglevel', 'INFO')
            self._logger.setLevel(loglevel)
        return self._logger

    @property
    def should_debug(self):
        return os.environ.get('DOCOPT_DEBUG')

    def apply_command_usage(self):
        # replace [usage:<command>] [options:<command>] placeholders with command.__doc__
        for command in self.commands:
            if command.__doc__:
                usage, options, description = command.get_command_doc()
                usage_placeholder = '[usage:{}]'.format(command.command)
                options_placeholder = '[options:{}]'.format(command.command)
                descr_placeholder = '[description:{}]'.format(command.command)
                self.docs = self.docs.replace(usage_placeholder, usage)
                self.docs = self.docs.replace(options_placeholder, options)
                self.docs = re.sub(r'\n\s*\n', '\n\n', self.docs)
            if any(v in self.argv for v in ('--help-ext', '-H', '-hh')):
                self.docs = self.docs.replace(descr_placeholder, description)
            else:
                self.docs = self.docs.replace(descr_placeholder, '')

    def parse(self):
        self.args = safe_docopt(self.docs, argv=self.argv, version=self.version)
        self.command = self.get_command()
        if not self.command:
            raise DocoptExit()
        if self.should_debug:
            print("*** docopt parsed args", self.args)
            print("*** docopt using command class {}".format(repr(self.command)))
            print("*** docopt using command method {}".format(repr(self.command.get_command_method())))
        return self

    def get_command(self):
        # find the command
        for commandcls in self.commands:
            global_command = False
            # given as literal
            command_asliteral = self.args.get(commandcls.command)
            # given as <command> arg
            command_asarg = self.args.get(self.command_tag) == commandcls.command
            # catch all?
            if not (command_asliteral or command_asarg):
                global_command = commandcls.command == 'global'
            if command_asliteral or command_asarg or global_command:
                command = commandcls(self.args, argv=self.argv,
                                     logger=self.logger, global_docs=self.docs)
                return command

    def process(self):
        # resolve command and method to call
        assert self.command is not None, "you must call DocoptParser.parse()"
        meth = self.command.get_command_method()
        if meth:
            return meth()


class DocoptCommand:
    """
    Base class for DocoptParser commands

    This assumes either of the following docopt formats:

        pgm.py <command> [<action>]
        pgm.py command action

    How to use:
        1. implement a DocoptCommand subclass for the command you wish to handle
        2. implement methods for each action
        3. to process command without action, implement __call__
        4. Pass your DocoptCommand classes as the command list to DocoptParser

        Each DocoptCommand subclass must at least provide the command name
        and either a valid action method or the __call__ method.

        Each action method, including __call__, is of format
            def action(self):
                # use self.args to get parsed arguments

        If a method with equal name as the action is found, it will be called.
        If no method is found, __call__ will be invoked, which by default raises
        NotImplementedError.

        Note a named <argument> will be used to lookup a method based on the
        value passed. e.g if <argument> is specified as foo then the foo method
        will be found.

        See DocoptParser for a complete example.

    Documentation of commands:
        A command class can have its own docstring in the same format as the
        main docstring, with its own Usage: and Options: sections. If it does
        so, docopt is automatically called on the command's docstring
        and self.args is updated accordingly.

        The command's Usage and Options sections can be included in the main
        docstring by adding [usage:<command>] and [options:<command>]
        placeholders. In this case the Usage: and Options: strings will be
        replaced by usage_header and options_header, respectively.

        See DocoptParser for a complete example.
    """
    command = 'unspecified'
    usage_header = 'Usage of {self.command}'
    options_header = 'Options for {self.command}'
    description_header = 'Working with {self.command}'

    def __init__(self, args, argv=None, logger=None, docs=None,
                 global_docs=None):
        # initialize, if this class contains a __doc__ header
        # with a Usage: section, self.args will automatically
        # be updated accordingly.
        self.args = args
        self.argv = argv
        self.logger = logger
        self.docs = docs or self.__doc__
        self.global_docs = global_docs
        self.parse()

    def parse(self):
        if (self.__doc__ and 'usage:' in self.__doc__.lower() or
                self.__doc__ and 'options:' in self.__doc__.lower()):
            docs = self.add_global_options()
            args = safe_docopt(docs, argv=self.argv)
            self.args.update(args)

    def add_global_options(self):
        # add Options: from top level parser. this is to make sure
        # that global options are checked on any [options] placeholder
        # in this command's docstring
        if self.global_docs and 'Options:' in self.global_docs:
            # get global options as clean as possible
            global_options = self.global_docs.split('Options:', 1)[1]
            global_options = global_options.split('\n')
            docs = dedent(self.docs).split('\n')
            # find options to add, stopping at the first empty line
            # after options. note the first line is always empty
            to_add = []
            for i, opt in enumerate(global_options):
                if i > 0 and not opt:
                    break
                if opt.strip().split('  ')[0] not in self.docs:
                    to_add.append(opt)
            docs.extend(to_add)

            def section(line):
                if line.lower().startswith('options'):
                    line = '\n' + line
                return line

            docs = '\n'.join(section(line) for line in docs)
        else:
            docs = self.docs
        return docs

    def __call__(self):
        raise NotImplementedError

    def get_command_method(self):
        # get the method to call, by default return command itself
        # for <action> style arguments will use the value to lookup
        # the method
        for k, v in self.args.items():
            lookup = str(v if k.startswith('<') else k)
            if v:
                meth = getattr(self, lookup, None)
                if meth is not None:
                    return meth
        return self

    def help(self):
        self.logger.info(dedent(self.__doc__).strip())

    @classmethod
    def get_command_doc(self):
        # return usage, options, description to replace
        # [usage:<command>], [options:<command>], [description:<command>]
        # -- format headers
        usage_header = self.usage_header.format(**locals())
        options_header = self.options_header.format(**locals())
        description_header = self.description_header.format(**locals())
        # get clean sections
        clean_command_doc = dedent(self.__doc__)
        clean_command_doc = clean_command_doc.replace('Usage:', 'usage:')
        clean_command_doc = clean_command_doc.replace('Options:', 'options:')
        clean_command_doc = clean_command_doc.replace('Description:', 'description:')
        # ensure all sections are present
        if 'options:' not in clean_command_doc:
            clean_command_doc += '\noptions:\n'
        if 'description:' not in clean_command_doc:
            clean_command_doc += '\ndescription:\n'
        # split
        usage, options = clean_command_doc.split('options:')
        options, description = options.split('description:')
        # ensure headers are not interfering with top-level
        usage = usage.replace('usage:', usage_header)
        options = (options_header + options) if options.strip() else ''
        description = (description_header + description) if description.strip() else ''
        return usage, options, description


def safe_docopt(doc, argv=None, help=True, version=None, options_first=False):
    try:
        args = docopt(doc, argv=argv, help=help, version=version, options_first=options_first)
    except DocoptLanguageError as e:
        print("*** ERROR {}, check below".format(e))
        print("arguments to doctopt were")
        print("   argv=", argv)
        print("   doc=", doc)
        print("**** ERROR {}, check above messages ***".format(e))
        exit(1)
    return args


def setup_console_logger():
    logger = logging.getLogger()
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



from omegaml.client.docoptparser import DocoptCommand


class GlobalCommand(DocoptCommand):
    """
    Usage: om <command> [<action>] [<args> ...] [options]
           om (models|datasets|scripts|jobs) [<args> ...] [--raw] [options]
           om runtime [<args> ...] [--async] [options]
           om cloud [<args> ...] [options]
           om [-h] [-hh] [--version] [--copyright]
    """
    command = "global"

    def __call__(self):
        if self.args.get('--copyright'):
            logger = self.logger
            logger.info("(c) omega|ml by one2seven GmbH, Zurich, Switzerland, https://omegaml.io")
            logger.info("third party components (c) by their respective copyright holders")
            logger.info("see LICENSE and THIRDPARTY-LICENSES at https://github.com/omegaml/omegaml/")
            return
        self.help()




from __future__ import absolute_import

import datetime
from mongoengine.base.fields import ObjectIdField
from mongoengine.document import Document
from mongoengine.fields import (
    StringField, FileField, DictField, DateTimeField
)

from omegaml.util import settings

# default kinds of objects
class MDREGISTRY:
    PANDAS_DFROWS = 'pandas.dfrows'  # dataframe
    PANDAS_SEROWS = 'pandas.serows'  # series
    PANDAS_HDF = 'pandas.hdf'
    PYTHON_DATA = 'python.data'
    PANDAS_DFGROUP = 'pandas.dfgroup'
    SKLEARN_JOBLIB = 'sklearn.joblib'
    OMEGAML_JOBS = 'script.ipynb'
    SPARK_MLLIB = 'spark.mllib'
    OMEGAML_RUNNING_JOBS = 'job.run'

    #: the list of accepted data types. extend using OmegaStore.register_backend
    KINDS = [
        PANDAS_DFROWS, PANDAS_SEROWS, PANDAS_HDF, PYTHON_DATA, SKLEARN_JOBLIB,
        PANDAS_DFGROUP, OMEGAML_JOBS, OMEGAML_RUNNING_JOBS, SPARK_MLLIB
    ]


def make_Metadata():
    # this is to create context specific Metadata classes that take the
    # database from the given alias at the time of use
    class Metadata(Document):
        """
        Metadata stores information about objects in OmegaStore
        """

        # fields
        #: this is the name of the data
        name = StringField(unique_with=['bucket', 'prefix'])
        #: bucket
        bucket = StringField()
        #: prefix
        prefix = StringField()
        #: kind of data
        kind = StringField(choices=MDREGISTRY.KINDS)
        #: for PANDAS_HDF and SKLEARN_JOBLIB this is the gridfile
        gridfile = FileField(
            db_alias='omega',
            collection_name=settings().OMEGA_MONGO_COLLECTION)
        #: for PANDAS_DFROWS this is the collection
        collection = StringField()
        #: for PYTHON_DATA this is the actual document
        objid = ObjectIdField()
        #: omegaml technical attributes, e.g. column indicies
        kind_meta = DictField()
        #: customer-defined other meta attributes
        attributes = DictField()
        #: s3file attributes
        s3file = DictField()
        #: location URI
        uri = StringField()
        #: created datetime
        created = DateTimeField(default=datetime.datetime.now)
        # the actual db is defined in settings, OMEGA_MONGO_URL
        meta = {
            'db_alias': 'omega',
            'indexes': [
                # unique entry
                {
                    'fields': ['bucket', 'prefix', 'name'],
                },
                'created',  # most recent is last, i.e. [-1]
            ]
        }

        def __eq__(self, other):
            return self.objid == other.objid

        def __unicode__(self):
            fields = ('name', 'bucket', 'prefix', 'created', 'kind')
            kwargs = ('%s=%s' % (k, getattr(self, k))
                      for k in self._fields.keys() if k in fields)
            return u"Metadata(%s)" % ','.join(kwargs)

    return Metadata


def make_QueryCache():
    class QueryCache(Document):
        collection = StringField()
        key = StringField()
        value = DictField()
        meta = {
            'db_alias': 'omega',
            'indexes': [
                'key',
            ]
        }

    return QueryCache



from omegaml.client.docoptparser import DocoptCommand
from omegaml.client.util import get_omega


class RuntimeCommand(DocoptCommand):
    """
    Usage:
      om runtime model <name> <model-action> [<X>] [<Y>] [<kw=value> ...>] [--async] [options]
      om runtime script <name> [<script-action>] [<args ...>] [--async] [options]
      om runtime job <name> [<job-action>] [<args ...>] [--async] [options]
      om runtime result <taskid> [options]
      om runtime ping [options]


    Options:
      --async  don't wait for results, will print taskid

    Description:
      <model-action> can be any valid model action like fit, predict, score,
      transform, decision_function etc.

      <script-action> defaults to run
      <job-action> defaults to run

      Examples:
        om runtime model <name> fit <X> <Y>
        om runtime model <name> predict <X>
    """
    command = 'runtime'

    def ping(self):
        om = get_omega(self.args)
        self.logger.info(om.runtime.ping())

    def _ensure_valid_XY(self, value):
        if value is not None:
            if value.isnumeric():
                return [eval(value)]
            if value[0] == '[' and value[-1] == ']':
                return eval(value)
        return value

    def model(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        action = self.args.get('<model-action>')
        async = self.args.get('--async')
        kwargs_lst = self.args.get('<kw=value>')
        X = self._ensure_valid_XY(self.args.get('<X>'))
        Y = self._ensure_valid_XY(self.args.get('<Y>'))
        # parse the list of kw=value values
        # e.g. key1=val1 key2=val2 => kwargs_lst = ['key1=val1', 'key2=val2']
        #   => kw_dct = { 'key1': eval('val1'), 'key2': eval('val2') }
        kv_dct = {}
        for kv in kwargs_lst:
            k, v = kv.split('=', 1)
            kv_dct[k] = eval(v)
        kwargs = {}
        if action == 'gridsearch':
            kwargs['parameters'] = kv_dct
        rt_model = om.runtime.model(name)
        meth = getattr(rt_model, action, None)
        if meth is not None:
            result = meth(X, Y, **kwargs)
            if not async:
                self.logger.info(result.get())
            else:
                self.logger.info(result)
            return
        raise ValueError('{action} is not applicable to {name}'.format(**locals()))

    def script(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        async = self.args.get('--async')
        result = om.runtime.script(name).run()
        if not async:
            self.logger.info(result.get())
        else:
            self.logger.info(result.task_id)

    def job(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        async = self.args.get('--async')
        result = om.runtime.job(name).run()
        if not async:
            self.logger.info(result.get())
        else:
            self.logger.info(result.task_id)

    def result(self):
        from celery.result import AsyncResult

        om = get_omega(self.args)
        task_id = self.args.get('<taskid>')
        result = AsyncResult(task_id, app=om.runtime.celeryapp).get()
        self.logger.info(result)
